//
// THIS CODE IS AUTO-GENERATED BY THE MOORE.IO CLIENT.  ANY CHANGES MADE WILL BE ERASED UPON THE NEXT BUILD.
//


// Filelists{% for filelist in filelists %}
-f {{ filelist }}{% endfor %}


// Defines{% for define in defines %}
{% if define.boolean %}+define+{{ define.name }}
{% else %}+define+{{ define.name }}={{ define.value }}{% endif %}{% endfor %}

// Directories{% for dir in dirs %}
+incdir+{{ dir }}{% endfor %}


// Files{% for file in files %}
{{ file }}{% endfor %}

